// =========  Start of Button.cpp generated code ==========
//  Button.cpp from Button by pfodGUIdesigner V1.0.4192
/*
  (c)2022 Forward Computing and Control Pty. Ltd.
  NSW Australia, www.forward.com.au
  This code is not warranted to be fit for any purpose. You may only use it at your own risk.
  This code may be freely used for both private and commercial use
  Provided this copyright is maintained.
*/

#include "Button.h"
#include <SafeString.h>

void Button::setCmdCallback(ButtonCallback cb) {
  buttonCallback = cb;
}

void Button::setVisible(bool _isVisible) { // flag true for visible false for hidden
   isVisible = _isVisible;
}

void Button::setTextColor(int color) {
  if (color < 0) {
    color = 0;
  }
  if (color > 255) {
    color = 255;
  }
  textColor = color;
}

void Button::setText(const char* text) {
  cSFA(sfText,buttonText);  
  sfText.clear();
  sfText.readFrom(text); // reads at most 10chars
}

Button::Button(pfodParser* parserPtr, pfodDwgs* dwgsPtr) : pfodDrawing(parserPtr, dwgsPtr) {
  forceReload = true; // force reload on reboot, e.g. reprogrammed
  buttonText[0] = '\0';
}

bool Button::processDwgCmds() { // return true if handled else false
  byte dwgCmd = parserPtr->parseDwgCmd(); // dwgCmd is '\0' if this is NOT a dwg cmd
  if (!dwgCmd) {
    return false; // not dwg cmd not handled
  }
  if (parserPtr->dwgCmdEquals(_touchZone_cmd_2)) { // handle _touchZone_cmd_..
//    Serial.print(" Got pfodAutoCmd:"); Serial.println("_touchZone_cmd_2");
//    printDwgCmdReceived(&Serial); // does nothing if passed NULL
    // add your cmd handling code here
    if (buttonCallback) {
      buttonCallback(); // this updates whole dwg
    } else {
      // send update
      sendUpdate();
    }
    return true;
  }
  return false; // not handled
}


void Button::drawDwg0() {
  dwgsPtr->index().idx(_update_idx_9).send();
  dwgsPtr->rectangle().idx(_idx_1).color(0).offset(0.00,0.00).size(23.00,8.00).rounded().centered().send();
  dwgsPtr->label().idx(_idx_2).color(textColor).offset(0.00,0.00).text(buttonText).fontSize(4).bold().center().send();
  dwgsPtr->touchZone().cmd(_touchZone_cmd_2).offset(0.00,0.00).size(23.00,8.00).centered().send();
  dwgsPtr->touchAction().cmd(_touchZone_cmd_2).action(
    dwgsPtr->hide().idx(_update_idx_9)
  ).send();
}

void Button::updateDwg() {
  dwgsPtr->rectangle().idx(_update_idx_9).color(254).offset(0.00,0.00).size(23.00,8.00).filled().rounded().centered().send();
  if (!isVisible) {
    dwgsPtr->hide().idx(_update_idx_9).send();
    dwgsPtr->hide().idx(_idx_1).send();
    dwgsPtr->hide().idx(_idx_2).send();
    dwgsPtr->hide().cmd(_touchZone_cmd_2).send();
  } else { // visible
    dwgsPtr->unhide().idx(_update_idx_9).send();
    dwgsPtr->unhide().idx(_idx_1).send();
    dwgsPtr->unhide().idx(_idx_2).send();
    dwgsPtr->unhide().cmd(_touchZone_cmd_2).send();
  }
}

bool Button::sendDwg() {
  if (!parserPtr->cmdEquals(*this)) {
    return false; // not this dwg's loadCmd
  }  // else
//  // ======== begin debugging
//  Serial.print(" sendDwg() cmd "); Serial.print((char*)parserPtr->getCmd());
//  if (parserPtr->isRefresh()) {
//    Serial.print(" isRefresh");
//  }
//  byte* argPtr = parserPtr->getFirstArg();
//  if (*argPtr) {
//    Serial.print("  args: ");
//    while (*argPtr) {
//      Serial.print("  "); Serial.print((char*)argPtr);
//      argPtr = parserPtr->getNextArg(argPtr);
//    }
//  }
//  Serial.print("\n");
//  //======= end debugging =======
  if (parserPtr->isRefresh() && !forceReload) { // refresh and not forceReload just send update
    sendUpdate();
  } else {
    forceReload = false; //
    long pfodLongRtn = 0;
    uint8_t* pfodFirstArg = parserPtr->getFirstArg(); // may point to \0 if no arguments in this msg.
    parserPtr->parseLong(pfodFirstArg, &pfodLongRtn); // parse first arg as a long
    sendFullDrawing(pfodLongRtn);
  }
  return true;
}

void Button::sendFullDrawing(long dwgIdx) {
  if (dwgIdx == 0) {
    dwgsPtr->start(50, 50, dwgsPtr->WHITE, true);  // the dwg size 50x50 and color (WHITE) are ignored when this drawing is inserted in the main dwg
    //true means more to come
    parserPtr->sendRefreshAndVersion(0); //need to set a version number for image refresh to work!!
    drawDwg0(); //
    dwgsPtr->end();
    
    // if drawing msg is >1023 split it in to two or more parts and add extra else if (dwgIdx == .. ) blocks here
    
  } else if (dwgIdx == 1) {
    dwgsPtr->startUpdate(false);  // always send update last so this is the last msg => false
    updateDwg(); // update with current state, // defined in display.cpp
    dwgsPtr->end();
    
  } else {
    parserPtr->print(F("{}")); // always return somthing
  }
}

void Button::sendUpdate() {
  dwgsPtr->startUpdate();  // dwg refresh updates must be <1024 bytes, i.e. can only be 1 msg
  updateDwg();
  dwgsPtr->end();
}

void Button::printDwgCmdReceived(Print *outPtr) {
  if (!outPtr) {
    return;
  }
  outPtr->print("     touchZone cmd "); outPtr->print((const char*)parserPtr->getDwgCmd()); outPtr->print(" at ");
  outPtr->print("("); outPtr->print(parserPtr->getTouchedX()); outPtr->print(",");
  outPtr->print(parserPtr->getTouchedY()); outPtr->print(") ");
  if (parserPtr->isTouch()) {
    outPtr->print("touch type:TOUCHED");
  }
  outPtr->println();
  if (*parserPtr->getEditedText()) { // first char not '\0'
    outPtr->print("     Edited text '"); outPtr->print((const char*)parserPtr->getEditedText()); outPtr->println("'");
  }
}
// =========  End of Button.cpp generated code ==========
