// For for XIAO esp32c3 with external led on D7 to 3V3 via 120R resistor
// This sketch controls the triac and accumulates the daily cost and hotwater kWhrs from the Watt readings from IoTaWatt

// Using ESP32 based board programmed via Arduino IDE
// follow the steps given on http://www.forward.com.au/pfod/ESP32/index.html to install ESP32 support for Arduino IDE
// NOTE: The latest versions of the ESP32 V2.0.0 - V2.0.5 Arduino board support have problems with WiFi connections
//  for the ESP32 chip (e.g. Sparkfun ESP32 Thing, tec).
// It may take 3 or more re-connections to get a stable connection after reboot or reprogramming.
// ESP32 BLE and Bluetooth are OK on V2.0.4 as is ESP32C3 WiFi (different chip).
// For ESP8266 V3.0.2 works.


// You need to modify the WLAN_SSID, WLAN_PASS settings below
// to match your network settings

/* Code generated by pfodDesignerV3 V3.0.4181
*/
/*
   (c)2014-2021 Forward Computing and Control Pty. Ltd.
   NSW Australia, www.forward.com.au
   This code is not warranted to be fit for any purpose. You may only use it at your own risk.
   This generated code may be freely used for both private and commercial use
   provided this copyright is maintained.
*/

#include <esp_task_wdt.h>
#include <pfodEEPROM.h>
#include <WiFi.h>
#include <WiFiClient.h>
#include <WebServer.h>
#include "millisDelay.h"
#include "locks.h"
#include "SafeString.h"
#include <pfodDwgs.h>
#include "Dwg_1.h"
#include "Button.h"
#include "Label.h"

#include <ESPAutoWiFiConfig.h>

const float IMPORT_COST_PER_kWhr = 0.3267;
const float EXPORT_COST_PER_kWhr = 0.07;
const float OFFPEAK_COST_PER_kWhr = 0.1596;

int ledPin = D7; // for XIAO esp32c3 external led
bool highForLedOn = false; // connected from 3v3 to D7
size_t eepromOffset = 32; // allow for security if you use EEPROM.begin(size) in your code add the size here so AutoWiFi data is written after your data


//#define DEBUG

// Level from 0-4
#define ASYNC_HTTP_DEBUG_PORT     Serial
#define _ASYNC_HTTP_LOGLEVEL_     0

#define HTTP_REQUEST_INTERVAL     5

const int ManualSwitchPin = D4; // set as input pullup, grounded by connection to D3 via manual switch
const int ManualSwitchGndPin = D3;  // set this output as low to provide gnd for D5.
const int TriacControlPin = D5; // D5
const float HeaterWattage = 3500;
const float HeaterHysteresisWatts = 500; // need -4000 before enable heater
const float HeaterSwitchOnValue = -(HeaterWattage + HeaterHysteresisWatts); // exports are -ve in 3PhaseWatt measurement
millisDelay boostTimer;
unsigned long BOOST_TIME_MS = 60l * 60 * 1000; // 1hr of boost

const unsigned MAX_HotWaterOffCount = 5; // 25sec delay before marking hotwater fully charged

int status;     // the Wifi radio's status

#include <WiFi.h>

// Download pfodESP32BufferedClient library from http://www.forward.com.au/pfod/pfodParserLibraries/index.html
// pfodESP32BufferedClient.zip contains pfodESP32BufferedClient and pfodESP32Utils
#include <pfodESP32Utils.h>
#include <pfodESP32BufferedClient.h>

// download the libraries from http://www.forward.com.au/pfod/pfodParserLibraries/index.html

// Seconds for timeout, default is 3s
// this must be above #include <AsyncHTTPRequest_Generic.h>
#define DEFAULT_RX_TIMEOUT           5
// To be included only in main(), .ino with setup() to avoid `Multiple Definitions` Linker Error
#include <AsyncHTTPRequest_Generic.h>             // https://github.com/khoih-prog/AsyncHTTPRequest_Generic


#include <pfodSecurity.h>
int swap01(int); // method prototype for slider end swaps

pfodSecurity parser("V2"); // create a parser with menu version string to handle the pfod messages
pfodDwgs dwgs(&parser);  // drawing support

pfodESP32BufferedClient bufferedClient;
const unsigned maxBadResponseCount = 4; // allow 4 failures => 20sec

const int portNo = 4989; // What TCP port to listen on for connections.

millisDelay restartBlockDelay;
const unsigned long RESTART_BLOCK_DELAY_MS = 2UL * 60 * 60 * 1000;  // 2ul* 60 * 60 * 1000;  //2hrs

void reboot() {
#ifdef DEBUG
  Serial.println("Called Reboot");
#endif
  if (restartBlockDelay.justFinished()) {
#ifdef DEBUG
    Serial.println("Rebooting...");
#endif
    hard_restart(); // loop here until wdt times out
  }
}

millisDelay totalCostUpdateTimer;
unsigned long TOTAL_COST_UPDATE_MS = 1UL * 60 * 1000; // update once/min
millisDelay yesterdayUpdateTimer; // if fails
unsigned long YESTERDAY_UPDATE_MS = 67 * 1000; // ~update once/min if fails

// add your pfod Password here for 128bit security
// eg "b0Ux9akSiwKkwCtcnjTnpWp" but generate your own key, "" means no pfod password
#define pfodSecurityCode ""
// see http://www.forward.com.au/pfod/ArduinoWiFi_simple_pfodDevice/index.html for more information and an example
// and QR image key generator.

// sfTime input 00:00 to 23:59, output 12:00am to 11:59pm
// input returned unchanged on errors
void convert24hrToAmPm(SafeString &sfTime) {
  cSF(sfStr, 10);
  cSF(sfMins, 10);
  if (sfTime.capacity() < 7) {
    return; // to small
  }
  if (sfTime.length() != 5) {
    return;
  }
  int idx = sfTime.indexOf(':');
  if (idx != 2) {
    return;
  }
  sfTime.substring(sfStr, 0, idx);
  sfTime.substring(sfMins, idx);
  bool isPM = false;
  int hrs = 0;
  if (sfStr.toInt(hrs)) {
    if (hrs >= 12) {
      hrs -= 12;
      isPM = true;
    }
    if (hrs == 0) {
      hrs += 12; // 00:.. => 12:.. am
    }
    sfTime.clear();
    sfTime.print(hrs);
    sfTime += sfMins;
    if (isPM) {
      sfTime += "pm";
    } else {
      sfTime += "am";
    }
  } else {
    return; // input unchanged
  }
}

// protected by timerFlagsLock
volatile bool requestSent = false;
volatile bool startResponseTimer = false;
volatile bool stopResponseTimer = false;
volatile bool startNextRequestTimer = false;

// protected by dataLock
volatile float ThreePhase_W = 0.0; // +ve for import -ve for export
volatile float SolarHW_W = 0.0; // +ve when heating hot water from solar
volatile float OffPeakHW_W = 0.0; // +ve when heating hot water from solar
volatile float Import_Whrs = 0.0; // +ve
volatile float Export_Whrs = 0.0; // -ve
volatile float SolarHW_Whrs = 0.0;
volatile float OffPeakHW_Whrs = 0.0;
volatile float YesterdaySolarHW_Whrs = 0.0;
volatile float YesterdayOffPeakHW_Whrs = 0.0;
volatile float YesterdayImport_Whrs = 0.0; // 0.0 means not set, if actualy value is 0.0 then it is changed to 0.0001 to be ne zero
volatile float YesterdayExport_Whrs = 0.0; // 0.0 means not set, if actualy value is 0.0 then it is changed to 0.0001 to be ne zero
volatile float YesterdayTotalCost = 0.0;
// when heating hot water from solar add back the amount to export as available for other use
// since solar hotwater will switch off in used elsewhere

// protected by stateLock
volatile bool HotWaterCanHeat = false; // true if power supplied to heater thermostat.
volatile bool HotWaterAuto = true;  // if false then HotWaterCanHeat is always set to true;
volatile bool WasHWfullyCharged = false; // true if hotwater full heated else false.
volatile unsigned RequestIdx = 0;// set if yesterdayTotalCost_cents == 0.0
const unsigned REQUEST_TOTALS = 1;
const unsigned REQUEST_YESTERDAY = 2; // else request watts

pfodAutoCmd mainDwgLoadCmd; // the main empty dwg
static bool forceMainMenuReload = true; // for testing force complete reload on every reboot
static bool forceMainDwgReload = true; // for testing force complete reload on every reboot
static unsigned long dwgRefresh = 5000;//set to 0 to disable auto refresh or 10000; to  refresh every 10sec
// when on the front screen, the mobile back button will also refresh, otherwise it goes back to the previous menu/screen

Dwg_1 _dwg_1(&parser, &dwgs); // create dwg object and add to parser processing
Button solarButton(&parser, &dwgs); // create dwg object and add to parser processing
Button boostButton(&parser, &dwgs); // create dwg object and add to parser processing
Label hotwaterkWhrs(&parser, &dwgs); // solar + offpeak totals
Label hotwaterState(&parser, &dwgs); // hot water either solar or boost
Label hotwaterFull(&parser, &dwgs); // either blank or Hot Water Fully Heated
Label totalCost(&parser, &dwgs); //
Label yesterday(&parser, &dwgs); //

// this must handle sending back response!!
// these call backs are on the loop() thread
void solarButtonCallback() {
  START_LOCK(stateLock) ;
  HotWaterAuto = true;
  HotWaterCanHeat = false;
  END_LOCK(stateLock);
  hotwaterState.setText("Solar Hot Water");
  hotwaterState.setTextColor(2);
  boostTimer.stop();
  sendMainMenuUpdate();
#ifdef DEBUG
  Serial.println(" set solar");
#endif
}

// this must handle sending back response!!
void boostButtonCallback() {
  START_LOCK(stateLock) ;
  HotWaterAuto = false;
  HotWaterCanHeat = false;
  END_LOCK(stateLock);
  hotwaterState.setText("Boosting Hot Water");
  hotwaterState.setTextColor(9);
  boostTimer.start(BOOST_TIME_MS); // start 1hr timer
  sendMainMenuUpdate();
#ifdef DEBUG
  Serial.println(" set mains");
#endif
}

Label nameTimeDisplay(&parser, &dwgs);
cSF(sfTime, 8); // 00:00 to 23:59 => 12:00am to 12:59pm

// send main dwg that just inserts the designed dwg at the pushZero position and scaling.
bool sendMainDwg() {
  // main dwg just inserts desgined dwg
  dwgs.start(50, 80, dwgs.WHITE, false); //false means NO more to come,  background defaults to WHITE if omitted i.e. dwgs.start(50,30);
  parser.sendRefreshAndVersion(0); //need to set a version number for image refresh to work!! send the parser version to cache this image, refreshes every 10sec
  dwgs.pushZero(25, 22, 1); // position and scale the inserted designed dwg
  dwgs.insertDwg().loadCmd(_dwg_1).send(); // insert the dwg object in the main dwg dwg
  dwgs.popZero();
  dwgs.pushZero(25, 46, 1); // position and scale the inserted designed dwg
  dwgs.insertDwg().loadCmd(totalCost).send(); // insert the dwg object in the main dwg dwg
  dwgs.popZero();
  dwgs.pushZero(25, 52, 1); // position and scale the inserted designed dwg
  dwgs.insertDwg().loadCmd(hotwaterkWhrs).send(); // insert the dwg object in the main dwg dwg
  dwgs.popZero();
  dwgs.pushZero(25, 56, 1); // position and scale the inserted designed dwg
  dwgs.insertDwg().loadCmd(hotwaterFull).send(); // insert the dwg object in the main dwg dwg
  dwgs.popZero();
  dwgs.pushZero(25, 61, 1); // position and scale the inserted designed dwg
  dwgs.insertDwg().loadCmd(yesterday).send(); // insert the dwg object in the main dwg dwg
  dwgs.popZero();
  dwgs.pushZero(13, 69, 0.8); // position and scale the inserted designed dwg
  dwgs.insertDwg().loadCmd(solarButton).send(); // insert the dwg object in the main dwg dwg
  dwgs.popZero();
  dwgs.pushZero(37, 69, 0.8); // position and scale the inserted designed dwg
  dwgs.insertDwg().loadCmd(boostButton).send(); // insert the dwg object in the main dwg dwg
  dwgs.popZero();
  dwgs.pushZero(25, 75, 1); // position and scale the inserted designed dwg
  dwgs.insertDwg().loadCmd(hotwaterState).send(); // insert the dwg object in the main dwg dwg
  dwgs.popZero();
  dwgs.end();
  return true;
}

void sendMainDwgUpdate() { // nothing here, but insertedDwgs updated by parser.parse()
  dwgs.startUpdate();  // dwg refresh updates must be <1024 bytes, i.e. can only be 1 msg
  dwgs.end();
}

void hard_restart() {
  esp_task_wdt_init(1, true); // watch dog timer times out in 1sec
  esp_task_wdt_add(NULL);
  while (true); // loop here until timeout force reboot.
}


AsyncHTTPRequest request;
millisDelay noResponseDelay;
const unsigned long NO_RESPONSE_DELAY_MS = 90ul * 1000; // 90sec CALL BACK not called, reboot must be >35sec

millisDelay requestDelay;
const unsigned long REQUEST_DELAY_MS = 5000; // min request is s-5s to s i.e. 5sec

WiFiServer server(portNo);
WiFiClient client;
WiFiClient* clientPtr; // non-null if have connected client

unsigned long plot_msOffset = 0; // set by {@} response
bool clearPlot = false; // set by the {@} response code

void sendRequest() {
  START_LOCK(timerFlagsLock) ;
  if (requestSent) {
    return;
  }
  END_LOCK(timerFlagsLock);

  float localYesterdayImport_Whrs = 0.0;
  START_LOCK(dataLock) ;
  localYesterdayImport_Whrs = YesterdayImport_Whrs;
  END_LOCK(dataLock);
  
  unsigned localRequestIdx = 0;
  START_LOCK(stateLock) ;
  localRequestIdx = RequestIdx;
  END_LOCK(stateLock);

  static bool requestOpenResult;
  if (request.readyState() == readyStateUnsent || request.readyState() == readyStateDone) {

    START_LOCK(stateLock) ;
    if (totalCostUpdateTimer.justFinished()) {
      localRequestIdx = REQUEST_TOTALS;
    }
    if (!totalCostUpdateTimer.isRunning()) {
      totalCostUpdateTimer.start(TOTAL_COST_UPDATE_MS);
    }
    if (yesterdayUpdateTimer.justFinished()) {
      if (localYesterdayImport_Whrs == 0.0) {
        localRequestIdx = REQUEST_YESTERDAY;
        yesterdayUpdateTimer.start(YESTERDAY_UPDATE_MS);
      }
    }
    RequestIdx = localRequestIdx;
    END_LOCK(stateLock);

    if (localRequestIdx == REQUEST_TOTALS) {
#ifdef DEBUG
      Serial.println("TotalCostRequest request");
#endif
      requestOpenResult = request.open("GET", "http://10.1.1.119/query?select=[Import.wh.d2,Export.wh.d2,Solar_HW.wh.d2,OffPeak_HW.wh.d2,time.local]&begin=d&end=s&group=all&format=csv&missing=skip");
      // ignore extra field
    } else if (localRequestIdx == REQUEST_YESTERDAY) { // this set when yestedayTotalCost_cents = 0.0 if request failes then reset after totalCost request
#ifdef DEBUG
      Serial.println("RequestYesterday request");
#endif
      requestOpenResult = request.open("GET", "http://10.1.1.119/query?select=[Import.wh.d2,Export.wh.d2,Solar_HW.wh.d2,OffPeak_HW.wh.d2,time.local]&begin=d-1d&end=d&group=all&format=csv&missing=skip");
    } else {
#ifdef DEBUG
      Serial.println("WattsRequest");
#endif
      requestOpenResult = request.open("GET", "http://10.1.1.119/query?select=[Watts_3Ph.d2,Watts_3Ph.d2,Solar_HW.d2,OffPeak_HW.d2,time.local]&begin=s-5s&end=s&group=all&format=csv&missing=skip");
    }
    if (requestOpenResult)   {
      // Only send() if open() returns true, or crash
      request.send();
      START_LOCK(timerFlagsLock) ;
      requestSent = true;
      startResponseTimer = true;
      END_LOCK(timerFlagsLock);
    } else {
      Serial.println(F("Can't send bad request"));
    }
  } else {
    Serial.println(F("Can't send request"));
  }
}

bool parseFloatToken(SafeString &sfToken, float &result) {
  bool rtn = true;
  // CHECK FOR 0
  sfToken.trim();
  if (sfToken.toFloat(result)) {
    // OK
  } else {
#ifdef DEBUG
    Serial.print(F("   ,")); Serial.print(sfToken); Serial.println(F(", is not a number"));
#endif
    rtn = false;
  }
  return rtn;
}

// returns false on error
bool parseResponse(SafeString &sfRespInit, float &watts_1, float &watts_2, float &hotwaterWatts, float &OffPeakHW_W) {

  unsigned localRequestIdx = 0;
  START_LOCK(stateLock) ;
  localRequestIdx = RequestIdx;
  END_LOCK(stateLock);

#ifdef DEBUG
  Serial.print(" response: "); Serial.println(sfRespInit);
#endif
  bool rtn = true;
  cSF(sfResp, sfRespInit.length());
  cSF(sfToken, sfRespInit.length());
  sfResp = sfRespInit.c_str();
  bool haveToken = sfResp.firstToken(sfToken, ','); // watts
  if (haveToken) {
    if (!parseFloatToken(sfToken, watts_1)) {
      rtn = false;
    }
  } else {
    //#ifdef DEBUG
    //    Serial.print("Bad response: "); Serial.println(sfRespInit);
    //#endif
    rtn = false;
  }

  haveToken = sfResp.nextToken(sfToken, ','); // hotwater
  if (haveToken) {
    if (!parseFloatToken(sfToken, watts_2)) {
      rtn = false;
    }
  } else {
    //#ifdef DEBUG
    //    Serial.print("Bad response: "); Serial.println(sfRespInit);
    //#endif
    rtn = false;
  }

  haveToken = sfResp.nextToken(sfToken, ','); // hotwater
  if (haveToken) {
    if (!parseFloatToken(sfToken, hotwaterWatts)) {
      rtn = false;
    }
  } else {
    //#ifdef DEBUG
    //    Serial.print("Bad response: "); Serial.println(sfRespInit);
    //#endif
    rtn = false;
  }

  haveToken = sfResp.nextToken(sfToken, ','); // offPeak_hotwater
  if (haveToken) {
    if (!parseFloatToken(sfToken, OffPeakHW_W)) {
      rtn = false;
    }
  } else {
    //#ifdef DEBUG
    //    Serial.print("Bad response: "); Serial.println(sfRespInit);
    //#endif
    rtn = false;
  }

  haveToken = sfResp.nextToken(sfToken, ','); // time
  if (haveToken) {
    sfToken.trim();
    // look for contains T00: if so reset kWhrs, i.e. reset between midnight and 1am each day
    START_LOCK(stateLock) ;
    if (localRequestIdx == 0) { // only watts has valid time
      if (sfToken.indexOf("T00:") > 0) { // reset kWhrs  this time seems to be OK even if zero returned for the watts etc
        reboot();
      }
      int timeIdx = sfToken.indexOf("T");
      if (timeIdx > 0) {
        sfToken.substring(sfTime, timeIdx + 1, timeIdx + 6);
        convert24hrToAmPm(sfTime);
      }
    }
    END_LOCK(stateLock);
  } else {
    //#ifdef DEBUG
    //    Serial.print("Bad response: "); Serial.println(sfRespInit);
    //#endif
    rtn = false;
  }

  haveToken = sfResp.nextToken(sfToken, ','); // unknown
  if (haveToken) {
#ifdef DEBUG
    Serial.print("Bad response (too many fields): "); Serial.println(sfRespInit);
#endif
    rtn = false;
  }

  // else
  return rtn;
}

void yesterdayRequestCB(void *optParm, AsyncHTTPRequest * request, int readyState) {
  static unsigned badResponseCount = 0;
  static unsigned int hotWaterOffCount = 0; // number of requests with hotwater < 100W and canHeat

  float localYesterdayImport_Whrs = 0.0;
  float localYesterdayExport_Whrs = 0.0;
  float localYesterdaySolarHW_Whrs = 0.0;
  float localYesterdayOffPeakHW_Whrs = 0.0;
  bool responseOK = true;
  (void) optParm;
  if (readyState != readyStateDone) {
    return;
  }
  //else

  if (request->responseHTTPcode() == 200) {
    cSF(sfResponse, request->available());
    request->responseRead((uint8_t*)(sfResponse.c_str()), sfResponse.capacity());
    // readResponse does not add terminating '\0' but SafeString does on next method call.
    responseOK = parseResponse(sfResponse, localYesterdayImport_Whrs, localYesterdayExport_Whrs, localYesterdaySolarHW_Whrs, localYesterdayOffPeakHW_Whrs); // parseResponse does not change sfResponse
    if (responseOK) {  // reset count
      badResponseCount = 0;
    } else {
      Serial.print("RequestYesterday Bad Data: "); Serial.println(sfResponse);
      badResponseCount++;
    }
  } else {
    Serial.println("RequestYesterday Bad response");
    badResponseCount++;
  }
#ifdef DEBUG
  Serial.print(" Response localYesterdayImport_Whrs:"); Serial.print(localYesterdayImport_Whrs);
  Serial.print(" localYesterdayExport_Whrs:"); Serial.print(localYesterdayExport_Whrs);
  Serial.print(" localYesterdaySolarHW_Whrs:"); Serial.print(localYesterdaySolarHW_Whrs);
  Serial.print(" localYesterdayOffPeakHW_Whrs:"); Serial.print(localYesterdayOffPeakHW_Whrs);
  Serial.print(" badResponseCount:"); Serial.print(badResponseCount);
  Serial.println();
#endif

  if ((badResponseCount == 0)) { // only update on response OK || (badResponseCount > maxBadResponseCount))) { // last one succeeded or have had more than 2 failures
    // if failures will update with 0.00
    // update accumulated SolarHW_Whrs
    START_LOCK(dataLock) ;
    if (localYesterdayImport_Whrs == 0.0) { // add as little to make it non-zero and avoid request
      localYesterdayImport_Whrs = 0.0001;
    }
    YesterdayImport_Whrs = localYesterdayImport_Whrs;
    YesterdayExport_Whrs = localYesterdayExport_Whrs;
    YesterdaySolarHW_Whrs = localYesterdaySolarHW_Whrs;
    YesterdayOffPeakHW_Whrs = localYesterdayOffPeakHW_Whrs;
    YesterdayTotalCost = (localYesterdayImport_Whrs / 1000.0) * IMPORT_COST_PER_kWhr
                         + (localYesterdayExport_Whrs / 1000.0) * EXPORT_COST_PER_kWhr
                         + (localYesterdayOffPeakHW_Whrs / 1000.0) * OFFPEAK_COST_PER_kWhr;

    END_LOCK(dataLock);
    START_LOCK(stateLock) ;
    RequestIdx = 0;
    END_LOCK(stateLock);
#ifdef DEBUG
    Serial.print("Updated ");
    Serial.print(" YesterdayTotalCost $:"); Serial.print(YesterdayTotalCost);
    Serial.print(" YesterdayImport_Whrs:"); Serial.print(YesterdayImport_Whrs);
    Serial.print("YesterdayExport_Whrs:"); Serial.print(YesterdayExport_Whrs);
    Serial.print(" YesterdaySolarHW_Whrs:"); Serial.print(YesterdaySolarHW_Whrs);
    Serial.print(" YesterdayOffPeakHW_Whrs:"); Serial.print(YesterdayOffPeakHW_Whrs); Serial.println();
#endif
  } else { // if (badResponseCount > maxBadResponseCount) { // max bad responsed skip this for now
    START_LOCK(stateLock) ;
    RequestIdx = 0;
    END_LOCK(stateLock);
    badResponseCount = 0;
  }

  START_LOCK(timerFlagsLock) ;
  //    if ((badResponseCount > 0) &&  (badResponseCount <= maxBadResponseCount)) { // last one succeeded or have had more than 2 failures
  //      resendRequest = true;
  //    }
  requestSent = false;
  startNextRequestTimer = true;
  stopResponseTimer = true;
  END_LOCK(timerFlagsLock);

}

//typedef enum {
//  readyStateUnsent      = 0,            // Client created, open not yet called
//  readyStateOpened      = 1,            // open() has been called, connected
//  readyStateHdrsRecvd   = 2,            // send() called, response headers available
//  readyStateLoading     = 3,            // receiving, partial data available
//  readyStateDone        = 4             // Request complete, all data available.
//} reqStates;

void totalCostRequestCB(void *optParm, AsyncHTTPRequest * request, int readyState) {
  static unsigned badResponseCount = 0;
  static unsigned int hotWaterOffCount = 0; // number of requests with hotwater < 100W and canHeat

  float localImport_Whrs = 0.0;
  float localExport_Whrs = 0.0;
  float localSolarHW_Whrs = 0.0;
  float localOffPeakHW_Whrs = 0.0;

  bool responseOK = true;
  (void) optParm;
  if (readyState != readyStateDone) {
    return;
  }
  //else

  if (request->responseHTTPcode() == 200) {
    cSF(sfResponse, request->available());
    request->responseRead((uint8_t*)(sfResponse.c_str()), sfResponse.capacity());
    // readResponse does not add terminating '\0' but SafeString does on next method call.
    responseOK = parseResponse(sfResponse, localImport_Whrs, localExport_Whrs, localSolarHW_Whrs, localOffPeakHW_Whrs); // parseResponse does not change sfResponse
    if (responseOK) {  // reset count
      badResponseCount = 0;
    } else {
      Serial.print("RequestTotals Bad Data: "); Serial.println(sfResponse);
      badResponseCount++;
    }
  } else {
    Serial.println("RequestTotals Bad response");
    badResponseCount++;
  }
#ifdef DEBUG
  Serial.print(" Response localImport_Whrs:"); Serial.print(localImport_Whrs);
  Serial.print(" localExport_Whrs:"); Serial.print(localExport_Whrs);
  Serial.print(" localSolarHW_Whrs:"); Serial.print(localSolarHW_Whrs);
  Serial.print(" localOffPeakHW_Whrs:"); Serial.print(localOffPeakHW_Whrs);
  Serial.print(" badResponseCount:"); Serial.print(badResponseCount);
  Serial.println();
#endif
  if (badResponseCount == 0) { // || (badResponseCount > maxBadResponseCount))) { // last one succeeded or have had more than 2 failures
    // if failures will update with 0.00
    // update accumulated SolarHW_Whrs
    START_LOCK(dataLock) ;
    Import_Whrs = localImport_Whrs;
    Export_Whrs = localExport_Whrs;
    SolarHW_Whrs = localSolarHW_Whrs;
    OffPeakHW_Whrs = localOffPeakHW_Whrs;
    END_LOCK(dataLock);

    START_LOCK(stateLock) ;
    RequestIdx = 0;
    END_LOCK(stateLock);
#ifdef DEBUG
    Serial.print("Updated TotalCost Import_Whrs:"); Serial.print(Import_Whrs);
    Serial.print(" Export_Whrs:"); Serial.print(Export_Whrs);
    Serial.print(" SolarHW_Whrs:"); Serial.print(SolarHW_Whrs);
    Serial.print(" OffPeakHW_Whrs:"); Serial.print(OffPeakHW_Whrs); Serial.println();
#endif
  } else if (badResponseCount > maxBadResponseCount) { // max bad responsed skip this for now
    START_LOCK(stateLock) ;
    RequestIdx = 0;
    END_LOCK(stateLock);
    badResponseCount = 0; // reset
  }

  START_LOCK(timerFlagsLock) ;
  //    if ((badResponseCount > 0) &&  (badResponseCount <= maxBadResponseCount)) { // last one succeeded or have had more than 2 failures
  //      resendRequest = true;
  //    }
  requestSent = false;
  startNextRequestTimer = true;
  stopResponseTimer = true;
  END_LOCK(timerFlagsLock);

}

void wattsRequestCB(void *optParm, AsyncHTTPRequest * request, int readyState) {
  static unsigned badResponseCount = 0;
  static unsigned int hotWaterOffCount = 0; // number of requests with hotwater < 100W and canHeat

  float watts3Ph = 0.0;
  float wattsDummy = 0.0;
  float hotwater = 0.0;
  float offPeak = 0.0;
  bool responseOK = true;
  (void) optParm;
  if (readyState != readyStateDone) {
    return;
  }
  //else
  if (request->responseHTTPcode() == 200) {
    cSF(sfResponse, request->available());
    request->responseRead((uint8_t*)(sfResponse.c_str()), sfResponse.capacity());
    // readResponse does not add terminating '\0' but SafeString does on next method call.
    responseOK = parseResponse(sfResponse, watts3Ph, wattsDummy, hotwater, offPeak); // parseResponse does not change sfResponse
    if (responseOK) {  // reset count
      badResponseCount = 0;
    } else {
      Serial.print("RequestWatts Bad Data: "); Serial.println(sfResponse);
      badResponseCount++;
    }
  } else {
    Serial.println("RequestWatts Bad response");
    badResponseCount++;
  }
#ifdef DEBUG
  Serial.print(" Response watts3Ph:"); Serial.print(watts3Ph); Serial.print(" hotwater:"); Serial.print(hotwater);
  Serial.print(" offPeak:"); Serial.print(offPeak);
  Serial.print(" badResponseCount:"); Serial.print(badResponseCount);   Serial.println();
#endif

  // local for use here
  bool localHotWaterCanHeat = false;
  bool localWasHWfullyCharged = false;
  START_LOCK(stateLock) ;
  localHotWaterCanHeat = HotWaterCanHeat;
  localWasHWfullyCharged = WasHWfullyCharged;
  END_LOCK(stateLock);

  if (!responseOK) {
    hotWaterOffCount = 0; // reset count
  } else {
    if (localHotWaterCanHeat) { // export >= 4kW
      if (hotwater < 100) { // hot water no current
        hotWaterOffCount++;
        if (hotWaterOffCount >= MAX_HotWaterOffCount) {  // can heat for 25sec and hot water no current for 25sec
          hotWaterOffCount = MAX_HotWaterOffCount; // limit it
          localWasHWfullyCharged = true; //set just once and leave set
        }
      } else {
        hotWaterOffCount = 0; // can heat and heating hot water reset count
      }
    } else {
      hotWaterOffCount = 0; // can not heat reset count
    }
  }

  if ((badResponseCount == 0 || (badResponseCount > maxBadResponseCount))) { // last one succeeded or have had more than 2 failures

    START_LOCK(dataLock) ;
    ThreePhase_W = watts3Ph;
    SolarHW_W = hotwater;
    END_LOCK(dataLock);
    badResponseCount = 0; // reset
#ifdef DEBUG
    Serial.print("Updated ThreePhase_W:"); Serial.print(ThreePhase_W); Serial.print(" SolarHW_W:"); Serial.print(SolarHW_W); Serial.println();
#endif
  }

  START_LOCK(stateLock) ;
  WasHWfullyCharged = localWasHWfullyCharged; // set above
  END_LOCK(stateLock);

  START_LOCK(timerFlagsLock) ;
  requestSent = false;
  startNextRequestTimer = true;
  stopResponseTimer = true;
  END_LOCK(timerFlagsLock);
}

void requestCB(void *optParm, AsyncHTTPRequest * request, int readyState) {
  unsigned localRequestIdx = 0;

  START_LOCK(stateLock) ;
  localRequestIdx = RequestIdx; // do yesterday totals now, take precedence over IsWattsRequest
  END_LOCK(stateLock);

  if (localRequestIdx == REQUEST_YESTERDAY) {
    yesterdayRequestCB(optParm, request, readyState);
  } else if (localRequestIdx == REQUEST_TOTALS) {
    totalCostRequestCB(optParm, request, readyState);
  } else {
    wattsRequestCB(optParm, request, readyState);
  }
}

void handle_pfodParser() {
  uint8_t cmd = parser.parse(); // parse incoming data from connection
  // parser returns non-zero when a pfod command is fully parsed and not consumed by an inserted dwg
  if (cmd != 0) { // have parsed a complete msg { to }
    if ('.' == cmd) {
      // pfodApp has connected and sent {.} , it is asking for the main menu
      if (parser.isRefresh() && !forceMainMenuReload) {
        sendMainMenuUpdate(); // menu is cached just send update
      } else {
        forceMainMenuReload = false;
        sendMainMenu(); // send back the menu designed
      }

    } else if (parser.cmdEquals(mainDwgLoadCmd)) {
      if (parser.isRefresh() && !forceMainDwgReload) { // refresh and not forceReload just send update
        sendMainDwgUpdate();
      } else {
        forceMainDwgReload = false; //
        sendMainDwg(); // just inserts dwgs and that will update them
      }

    } else if (parser.cmdEquals('A')) { // click in A menu item that holds the main dwg
      // not used as inserted dwgs handle all the clicks
      sendMainMenuUpdate(); // always reply to the msg.

      // handle {@} request
    } else if ('@' == cmd) { // pfodApp requested 'current' time
      parser.print(F("{@`0}")); // return `0 as 'current' raw data milliseconds
    } else if ('!' == cmd) {
      // CloseConnection command
      closeConnection(parser.getPfodAppStream());
    } else {
      // unknown command
      parser.print(F("{}")); // always send back a pfod msg otherwise pfodApp will disconnect.
    }
  }
}

void setup() {
  EEPROM.begin(512);  // only use 20bytes for pfodSecurity but reserve 512 (pfodWifiConfig uses more)

  Serial.begin(115200); // set Serial for error msgs
#ifdef DEBUG
  for (int i = 10; i > 0; i--) {
    Serial.print(i); Serial.print(' ');
    delay(500);
  }
  Serial.println();
#endif

  if (initializeLocks()) {
    Serial.println("Error: could not create locks");
    hard_restart();
  }

  pinMode(TriacControlPin, OUTPUT);
  digitalWrite(TriacControlPin, LOW);
  pinMode(ManualSwitchGndPin, OUTPUT);
  digitalWrite(ManualSwitchGndPin, LOW); // a GND for manual switch
  pinMode(ManualSwitchPin, INPUT_PULLUP); // pulled low by manual switch to ManualSwitchGndPin when in Solar position

  restartBlockDelay.start(RESTART_BLOCK_DELAY_MS);
  totalCostUpdateTimer.start(15000); // restart timer
  yesterdayUpdateTimer.start(1000);

  // set initial display state
  hotwaterkWhrs.setTextColor(0);
  hotwaterkWhrs.setTextSize(4);
  hotwaterFull.setTextSize(3);
  hotwaterFull.setTextColor(2);
  totalCost.setTextSize(6);
  yesterday.setTextSize(2);
  nameTimeDisplay.setTextSize(4);
  nameTimeDisplay.setTextColor(12);
  hotwaterState.setTextColor(2);
  hotwaterState.setTextSize(4);


  // Explicitly set the ESP32 to be a WiFi-client, otherwise, it by default,
  // would try to act as both a client and an access-point and could cause
  // network-issues with your other WiFi-devices on your WiFi-network.
  WiFi.mode(WIFI_STA);
#ifdef DEBUG
  setESPAutoWiFiConfigDebugOut(Serial); // turns on debug output for the ESPAutoWiFiConfig code
#endif

  if (ESPAutoWiFiConfigSetup(ledPin, highForLedOn, 0)) { // check if we should start access point to configure WiFi settings
    return; // in config mode so skip rest of setup
  }

  solarButton.setText("<b><+2>Solar"); // bold font size +2 (+6 is double)
  solarButton.setTextColor(2);
  solarButton.setCmdCallback(&solarButtonCallback);
  boostButton.setText("Boost 1hr");
  boostButton.setTextColor(1);
  boostButton.setCmdCallback(&boostButtonCallback);
  // set initial display state
  hotwaterkWhrs.setTextColor(0);
  hotwaterkWhrs.setTextSize(4);
  hotwaterkWhrs.setText("Hot Water ");
  // set initial display state
  if (HotWaterAuto) {
    hotwaterState.setText("Solar Hot Water");
    hotwaterState.setTextColor(2);
  } else {
    hotwaterState.setText("Boosting Hot Water");
    hotwaterState.setTextColor(9);
  }
  totalCost.setTextSize(6);
  yesterday.setTextSize(2);

#ifdef DEBUG
  Serial.println();
  Serial.println("WiFi connected");
#endif

  // Start the server
  server.begin();
#ifdef DEBUG
  Serial.println("Server started");
#endif

  // Print the IP address
#ifdef DEBUG
  Serial.println(WiFi.localIP());
#endif

  request.setDebug(false);
  request.onReadyStateChange(requestCB);
  requestDelay.start(100); // first request in 100ms

}

void loop() {
  //  .. other code that MUST run all the time
  if (ESPAutoWiFiConfigLoop()) {  // handle WiFi config webpages
    return;  // skip the rest of the loop until config finished
  }


  // local for use here
  float localThreePhase_W = 0.0;  // +ve for import -ve export
  float localSolarHW_W = 0.0; // +ve for heating
  START_LOCK(dataLock) ;
  localThreePhase_W = ThreePhase_W;
  localSolarHW_W = SolarHW_W;
  END_LOCK(dataLock);

  // local for use here
  bool localHotWaterCanHeat = false;
  bool localHotWaterAuto = false;
  START_LOCK(stateLock) ;
  localHotWaterCanHeat = HotWaterCanHeat;
  localHotWaterAuto = HotWaterAuto;
  END_LOCK(stateLock);
  if (boostTimer.justFinished()) {
    localHotWaterAuto = true; // back to solar
    localHotWaterCanHeat = false;
  }

  if (!localHotWaterAuto) {
    localHotWaterCanHeat = true; // can always heat on mains power
  } else {
    if (localHotWaterCanHeat) {
      if (localThreePhase_W >= 0.0) {
        localHotWaterCanHeat = false; // not enought power to run heater
      }
    } else { // not running localHotWaterCanHeat = false;
      if (localThreePhase_W < HeaterSwitchOnValue) { // -4000
        localHotWaterCanHeat = true;
      }
    }
  }

  if (localHotWaterCanHeat) { // turn on triac
    digitalWrite(TriacControlPin, HIGH);
  } else {
    digitalWrite(TriacControlPin, LOW);
  }

  START_LOCK(stateLock) ;
  HotWaterCanHeat = localHotWaterCanHeat;
  HotWaterAuto = localHotWaterAuto;
  END_LOCK(stateLock);

  // check for hotwater fully charged
  // if localHotWaterCanHeat && localSolarHW_W <100 (to allow for noise) for > 15sec then switch to full
  // if request fails restart timer
  // if localHotWaterCanHeat && localSolarHW_W > 2500 then switch to not full

  if (requestDelay.justFinished()) {
    sendRequest();
  }
  bool rebootFlag = false;

  START_LOCK(timerFlagsLock) ;
  if (startNextRequestTimer) {
    startNextRequestTimer = false;
    requestDelay.start(REQUEST_DELAY_MS);
  }

  if (startResponseTimer) {
    startResponseTimer = false;
    noResponseDelay.start(NO_RESPONSE_DELAY_MS);
  }
  if (stopResponseTimer) {
    stopResponseTimer = false;
    noResponseDelay.stop();
  } else {
    if (noResponseDelay.justFinished()) {
      rebootFlag = true;
    }
  }
  END_LOCK(timerFlagsLock);
  if (rebootFlag) {
    Serial.println("No response reboot");
    hard_restart(); // force WDT to timeout // could also use ESP.restart()see https://github.com/esp8266/Arduino/issues/1017  seems to work here
  }
  _dwg_1.updateDwg_1(localThreePhase_W, localSolarHW_W, localHotWaterAuto);
  handle_pfodServerConnection();  // may update state
}

void handle_pfodServerConnection() {
  if (clientPtr && (!clientPtr->connected())) {
    closeConnection(parser.getPfodAppStream());
  }
  if (server.hasClient()) { // new connection
    if (!clientPtr) { // no existing connection
      client = server.available(); // get any new client
      clientPtr = &client;
      // have connected client
      parser.connect(bufferedClient.connect(clientPtr), F(pfodSecurityCode)); // sets new io stream to read from and write to
      EEPROM.commit(); // does nothing if nothing to do
    } else { // already have client so just stop the new one
      WiFiClient newClient = server.available(); // get any new client
      newClient.stop();
    }
  }
  if (clientPtr) {
    handle_pfodParser();
  }
}

void closeConnection(Stream * io) {
  (void)(io); // unused
  // add any special code here to force connection to be dropped
  parser.closeConnection(); // nulls io stream
  bufferedClient.stop(); // clears client reference
  clientPtr->stop();
  clientPtr = NULL;
}

void setLable() {
  bool manualSwitchOnSolar = (digitalRead(ManualSwitchPin) == LOW); // true if low else false

  float localTotalCost = 0.0;
  float localYesterdayTotalCost = 0.0;

  float localImport_Whrs = 0.0;
  float localExport_Whrs = 0.0;
  float localSolarHW_Whrs = 0.0;
  float localOffPeakHW_Whrs = 0.0;

  float localYesterdayImport_Whrs = 0.0;
  float localYesterdayExport_Whrs = 0.0;
  float localYesterdaySolarHW_Whrs = 0.0;
  float localYesterdayOffPeakHW_Whrs = 0.0;

  START_LOCK(dataLock) ;
  localImport_Whrs = Import_Whrs;
  localExport_Whrs = Export_Whrs;
  localSolarHW_Whrs = SolarHW_Whrs;
  localOffPeakHW_Whrs = OffPeakHW_Whrs;
  localYesterdaySolarHW_Whrs = YesterdaySolarHW_Whrs;
  localYesterdayOffPeakHW_Whrs = YesterdayOffPeakHW_Whrs;
  localYesterdayTotalCost = YesterdayTotalCost;
  END_LOCK(dataLock);

  cSF(sfLabel, 45);
  bool localHotWaterAuto = true;
  bool localWasHWfullyCharged = false;
  START_LOCK(stateLock) ;
  localHotWaterAuto = HotWaterAuto;
  localWasHWfullyCharged = WasHWfullyCharged;
  END_LOCK(stateLock);

  sfLabel.clear();
  sfLabel.print("Cost today ");
  localTotalCost = (localImport_Whrs / 1000.0) * IMPORT_COST_PER_kWhr
                   + (localExport_Whrs / 1000.0) * EXPORT_COST_PER_kWhr
                   + (localOffPeakHW_Whrs / 1000.0) * OFFPEAK_COST_PER_kWhr;
  if ((localTotalCost < 0.01) && (localTotalCost > -0.01)) {  // ~zero
    totalCost.setTextColor(0);
    localTotalCost = 0.0;
  } else if (localTotalCost > 0) {
    totalCost.setTextColor(9);
  } else {
    totalCost.setTextColor(2);
    sfLabel.print("-");
    localTotalCost = -localTotalCost;
  }
  sfLabel.print("$");
  sfLabel.print(localTotalCost, 2);
  totalCost.setText(sfLabel.c_str());

  sfLabel.clear();
  sfLabel.print("Yesterday  ");
  if ((localYesterdayTotalCost < 0.01) && (localYesterdayTotalCost > -0.01)) {  // ~zero
    localYesterdayTotalCost = 0.0;
  } else if (localYesterdayTotalCost > 0) {
  } else {

    sfLabel.print("-");
    localYesterdayTotalCost = -localYesterdayTotalCost;
  }
  sfLabel.print("$");
  sfLabel.print(localYesterdayTotalCost, 2);

  sfLabel.print("  ");
  sfLabel.print((localYesterdaySolarHW_Whrs + localYesterdayOffPeakHW_Whrs) / 1000.0, 1);
  sfLabel.print("kWhr");
  yesterday.setText(sfLabel.c_str());

  sfLabel.clear();
  sfLabel.print("Hot Water ");
  sfLabel.print((localSolarHW_Whrs + localOffPeakHW_Whrs) / 1000.0, 1);
  sfLabel.print("kWhr");
  hotwaterkWhrs.setText(sfLabel.c_str());
  if (localWasHWfullyCharged) {
    hotwaterFull.setText("Hot Water was Fully Heated");
  } else {
    hotwaterFull.setText("");
  }
  sfLabel.clear();
  if (manualSwitchOnSolar) {
    solarButton.setVisible(true);
    boostButton.setVisible(true);
    if (HotWaterAuto) {
      sfLabel.prefix("Solar Hot Water");
      hotwaterState.setText(sfLabel.c_str());
      hotwaterState.setTextColor(2);
    } else {
      sfLabel.prefix("Boosting Hot Water");
      hotwaterState.setText(sfLabel.c_str());
      hotwaterState.setTextColor(9);
    }
    hotwaterState.setTextSize(4);
  } else {
    solarButton.setVisible(false);
    boostButton.setVisible(false);
    sfLabel.print("Hot Water Off Peak\nor OFF");
    hotwaterState.setText(sfLabel.c_str());
    hotwaterState.setTextColor(0);
    hotwaterState.setTextSize(4);
  }
}

void sendMainMenu() {
  setLable();
  parser.print(F("{,"));  // start a Menu screen pfod message
  parser.print("<bg w>~"); // no prompt
  parser.sendRefreshAndVersion(dwgRefresh);
  // send menu items
  parser.print(F("|+A~"));
  parser.print(mainDwgLoadCmd); // the unique load cmd
  parser.print(F("}"));  // close pfod message
}

void sendMainMenuUpdate() {
  setLable();
  parser.print(F("{;"));  // start an Update Menu pfod message
  parser.print(F("~")); // no change to colours and size
  // send menu items
  parser.print(F("|+A")); // reload A calls for dwg z update
  parser.print(F("}"));  // close pfod message
  // ============ end of menu ===========
}


int swap01(int in) {
  return (in == 0) ? 1 : 0;
}
// ============= end generated code =========
