/*
   locks.cpp
   by Matthew Ford,  2019/11/16
   (c)2019 Forward Computing and Control Pty. Ltd.
   NSW, Australia  www.forward.com.au

*/
// note ESP32 by default does NOT support static Semaphores

#include "locks.h"

/*  volatile alternative to locks

volatile unsigned char *my_data;
This declares my_data to be a pointer to volatile unsigned char.

To make the pointer itself volatile, you'd need this instead:
unsigned char *volatile my_data;

And of course, both the pointer and the pointed-to data may be volatile:
volatile unsigned char *volatile my_data;
*/

/** a lock to control access to timer control flags */
SemaphoreHandle_t timerFlagsLock;
/** a lock to control access to data i.e. power measurements */
SemaphoreHandle_t dataLock;
/** a lock to control access to system state flags, e.g. Auto Manual, On/Off*/
SemaphoreHandle_t stateLock;


/**
  creates all the locks needed.
  returns non-zero on error<br>
  can be called more than once
*/
int initializeLocks() {
  if (timerFlagsLock == NULL) {
    timerFlagsLock = xSemaphoreCreateRecursiveMutex();
    if (timerFlagsLock == NULL) {
      return 1;
    }
  }
  if (dataLock == NULL) {
    dataLock = xSemaphoreCreateRecursiveMutex();
    if (dataLock == NULL) {
      return 1;
    }
  }
  if (stateLock == NULL) {
    stateLock = xSemaphoreCreateRecursiveMutex();
    if (stateLock == NULL) {
      return 1;
    }
  }
  return 0;
}
